/*
 * $RCSfile: MachineId.cc,v $
 * $Revision: 1.4 $
 *
 * Comments:
 *
 * (C) Copyright ParaSoft Corporation 2008.  All rights reserved.
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft
 * The copyright notice above does not evidence any
 * actual or intended publication of such source code.
 *
 * $Author: mlove $          $Locker:  $
 * $Date: 2008-04-15 01:33:11 $
 * $Log: not supported by cvs2svn $
 * Revision 1.3  2008/04/13 18:57:52  mlove
 * changed to base machine id on hash of Mac serial number (PR 56775)
 *
 * Revision 1.2  2008/04/11 21:13:41  mlove
 * added error message to help with troubleshooting (PR56775)
 *
 * Revision 1.1  2008/04/11 16:06:52  mlove
 * initial implementation of license checking on Mac OS X (PR 56775)
 *
 */

#include "MachineId.h"
#include "../Error.h"
#include <IOKit/IOKitLib.h>
#include <CoreFoundation/CoreFoundation.h>

int MachineId::getMachineId()
{
    int result = 0;
    io_service_t  pexpert;
    CFStringRef   serial;
   
    // get the Platform Expert object
    pexpert = IOServiceGetMatchingService(kIOMasterPortDefault,
                  IOServiceMatching("IOPlatformExpertDevice"));
    if (!pexpert) {
        Error::setErrorMessage("Unable to get IOPlatformExpertDevice from Mac OS X kernel");
        return 0;
    }
   
    serial = (CFStringRef)IORegistryEntryCreateCFProperty(
                 pexpert, CFSTR(kIOPlatformSerialNumberKey),
                 kCFAllocatorDefault,
                 kNilOptions);
    if (serial) {
        // CFShow(serial); // note that this will go to stderr
        result = CFHash(serial);
        CFRelease(serial);
    } else {
        Error::setErrorMessage("Unable to get the Mac serial number from IOPlatformExpertDevice");
    }
   
    IOObjectRelease(pexpert);
   
    return result;
}


